﻿using System.Collections;
using System.Collections.Generic;
using UnityEngine;
using UnityEngine.UI;

using System;
using System.IO;
using System.IO.Ports;
using System.Linq;
using System.Runtime.InteropServices;
using System.Threading;


// 16 bits in a byte
// Should be 8 bytes
// Reversed data
[StructLayout(LayoutKind.Sequential, Pack = 0)]
public struct Acc_Struct {
    public short x;
    public short y;
    public short z;
}

[StructLayout(LayoutKind.Sequential, Pack = 0)]
public struct ArduinoData {
    public Acc_Struct acc_struct;         // accelerometer struct
    public ushort pot_val;                // potentiometer0 val
    public ushort servo_val;                // potentiometer0 val
};

public class ThreadSerialCommunication : MonoBehaviour
{
    public static string portName = "COM8";
    public static int baudrate = 115200;
    public bool looping = true;
    // Opens the connection on the serial port
    SerialPort stream = new SerialPort(portName, baudrate);

    float timer;

    bool continueReading;
    bool continueWriting;

    int num_bytes_read = 0;
    byte checksum = 0;

    private const int buffLen = 2560;
    private const int header = 4863;  // Convert.ToInt32(0x12FF);
    private const int endbyte = 13550; //Convert.ToInt32(0x34EE);

    int data_sz = Marshal.SizeOf(typeof(ArduinoData));
    byte[] buff = new byte[buffLen];

    public ArduinoData[] arduinoData = new ArduinoData[buffLen];
    public ArduinoData latestArduinoData;
    int arduinoData_counter = 0;
    int arduinoData_last_update = 0;

    /* Threading Variables */
    private Thread thread;
    // When ready, thread reads input queue then when ready outputs to output
    // queue
    private Queue outputQueue;
    private Queue inputQueue;

    void Start () {
        stream.ReadTimeout = 50;
        stream.Open();
        StartThread();
    }

	void OnApplicationQuit() {
        StopThread();
    }

    public void StartThread() {
        outputQueue = Queue.Synchronized(new Queue());
        inputQueue = Queue.Synchronized(new Queue());

        thread = new Thread(ThreadLoop);
        thread.Start();
    }

    public void ThreadLoop() {
        // Looping
        string command;
		DateTime initialTime = DateTime.Now;
		DateTime nowTime;
		TimeSpan diff = default(TimeSpan);
		long dt = 0;
        while (IsLooping()) {
            // Check this every 20ms. 50Hz, that's fast enough 
            // Send to Arduino

            /* if (dt >= 20) { */
            /*     outputQueue.Enqueue("T1"); */
            /*     dt = 0; */
            /* } */
            if (outputQueue.Count != 0)
            {
                command = (string)outputQueue.Dequeue();
                if (command.Contains("T")) {
                    Debug.Log("sending to arduino: " + command);
                }
                WriteToArduino(command);
            }

            // Read from Arduino
            byte[] result = ReadFromArduino();
            if (result != null) {
                inputQueue.Enqueue(result);
                //Debug.Log("before processing");
                /* Debug.Log(BitConverter.ToString(buff)); */
            }
            // Restrict memory use
            if (inputQueue.Count > 0) {
                result = (byte[])inputQueue.Dequeue();
                ProcessCommInfo(result, result.Length);
                ResetBuffer();
            }
			nowTime = DateTime.Now;
			diff = nowTime - initialTime;
            dt += diff.Milliseconds;
        }
        stream.Close();
        Debug.Log("Closed port");
    }

    public void WriteToArduino(string command) {
        stream.WriteLine(command);
        stream.BaseStream.Flush();
    }

    public byte[] ReadFromArduino (int timeout = 0) {
        stream.ReadTimeout = timeout;        
        try {
            int bytesToRead = stream.BytesToRead;
            if (bytesToRead > 0) {
                StreamToBuffer(bytesToRead);
            }
            return buff;
        }
        catch (TimeoutException e) {
            return null;
        }
    }

    public byte[] GetLatestBuffer () {
        byte[] buffer;
        lock (this) {
            buffer = (byte[])inputQueue.Dequeue();
            return buffer;
        }
    }

	void StreamToBuffer(int bytesToRead) {
        int b;
        int num_bytes = 0;

        while(num_bytes < bytesToRead) {
            b = stream.ReadByte();
            buff[num_bytes_read] = Convert.ToByte(b);
            num_bytes_read++;
            num_bytes++;
            if (num_bytes_read > buff.Length - 8) {
                ResetBuffer();
            }
        }
		stream.DiscardInBuffer();
		stream.Write ("T0".ToCharArray(), 0, 2);
	}

	void ProcessCommInfo(byte[] comm_buff, int bytesToRead) {
		int checkbyte = 0;
		int priorbyte = comm_buff[0];
        int b;

        // Copy the buffer into another space to avoid overriding during
        // processing from coroutine / thread
        byte[] buff_copy = comm_buff;
		byte checksum_test = 0;
        byte [] data = new byte[data_sz];

        int hdr_ind = 0;
        int edr_ind = 0;

        for (int i = 1; i < bytesToRead; i++) {
            b = buff_copy[i];
            if (priorbyte == 255 && b == 18) {
                b = Read16BitValue(priorbyte,b);
            } else if (priorbyte == 238 && b == 52) {
                b = Read16BitValue(priorbyte, b);
            }
            if (b == header) {
                hdr_ind = i;
            } else if (b == endbyte) {
                edr_ind = i;
                if (edr_ind < 2) {
                    continue;
                }
                Array.Copy(buff_copy, hdr_ind+1, data, 0, data_sz);
                checksum_test = CalculateChecksum(data,  data.Length);
                checksum = (byte)buff_copy[edr_ind - 2];
                /* Debug.Log("chksum test: " + checksum_test + "\tchecksum rcx: " + checksum); */
                arduinoData_last_update = arduinoData_counter;
                if (checksum == checksum_test) {
                    arduinoData[arduinoData_counter] = ByteArrayToStruct (data, 
                            arduinoData[arduinoData_counter], 0);
                    latestArduinoData = arduinoData[arduinoData_counter];
                    arduinoData_counter++;
                    if (arduinoData_counter >= arduinoData.Length) {
                        arduinoData_counter = 0;
                    }
                } else {
                    Debug.LogWarning ("Checksum does not equate");
                }
            }
            priorbyte = b;
        }
	}

    public void StopThread() {
        lock (this) {
            looping = false;
            stream.DiscardInBuffer();
            stream.Write ("T0".ToCharArray(), 0, 2);
            stream.Close();
            Debug.Log("Closed port");
        }
    }

    public bool IsLooping() {
        lock (this) {
            return looping;
        }
    }

	private byte CalculateChecksum(byte[] dataToCalculate, int length) {
		int chksum = 0;
		for (int i = 0; i < length; i++) {
			chksum += dataToCalculate[i]; // increment pointer and de-reference the unincremented address
		}
		chksum &= 0xff;
		return (byte)chksum;
	}

	private int Read16BitValue(int b1, int b2) {
		int b2_shift = b2 << 8;
		return (b1 + b2_shift);
	}

	ArduinoData ByteArrayToStruct(byte[] bytearray, ArduinoData structureObj, int position) {
		int length = Marshal.SizeOf (typeof(ArduinoData));
		IntPtr ptr = Marshal.AllocHGlobal (length);
		Marshal.Copy (bytearray, 0, ptr, length);

		structureObj = (ArduinoData)Marshal.PtrToStructure (ptr, structureObj.GetType());
		//structureObj = (ArduinoData)Marshal.PtrToStructure(Marshal.UnsafeAddrOfPinnedArrayElement(bytearray,
		//	position), structureObj.GetType());

		Marshal.FreeHGlobal (ptr);
		return structureObj;
	}

    public int GetArduinoDataCounter() {
        return arduinoData_counter;
    }

    public ArduinoData[] GetArduinoData() {
        return arduinoData;
    }

    private void ResetBuffer() {
        for (int i = 0; i < buffLen; i++) {
            buff[i] = 0x00;
        }
        num_bytes_read = 0;
    }

    public void EnqueueOutput(string command) {
        outputQueue.Enqueue(command);
    }
}
